"""Holds functions to prepare the strings and text in the dataframe."""
from typing import Any, Optional, Sequence, Union

import numpy as np
import pandas as pd

from forestplot.arg_validators import check_iterables_samelen
from forestplot.dataframe_utils import insert_empty_row


def form_est_ci(
    dataframe: pd.core.frame.DataFrame,
    estimate: str,
    ll: str,
    hl: str,
    decimal_precision: int,
    caps: Union[tuple, list, str] = "()",
    connector: str = " to ",
) -> pd.core.frame.DataFrame:
    """Form the estimated effect sizes and corresponding confidence intervals as a formatted column.

    Creating only on demand.

    Parameters
    ----------
    dataframe (pandas.core.frame.DataFrame)
            Pandas DataFrame where rows are variables. Columns are variable name, estimates,
            margin of error, etc.
    estimate (str)
            Name of column containing the estimates (e.g. pearson correlation coefficient,
            OR, regression estimates, etc.).
    ll (str)
            Name of column containing the lower limit of the confidence intervals.
    hl (str)
            Name of column containing the upper limit of the confidence intervals.
    decimal_precision (int)
            Precision of 2 means we go from '0.1234' -> '0.12'.
    caps (iterable)
            Eg '()' means that confidence intervals are enclosed in brackets (eg (-0.1 to 0.4)).
    connector (str)
            How to connect lower and upper limits of confidence intervals. Eg ' to ' means
            confidence intervals are of the form 'll to hl'.

    Helpers
    -------
            _right_justify_num

    Returns
    -------
            pd.core.frame.DataFrame with an additional formatted 'est_ci' column.
    """
    if ll is None:
        cols = [estimate]
    else:
        cols = [estimate, ll, hl]
    for col in cols:
        dataframe = _right_justify_num(
            dataframe=dataframe, col=col, decimal_precision=decimal_precision
        )
    for ix, row in dataframe.iterrows():
        formatted_est = row[f"formatted_{estimate}"]
        if ll is not None:
            formatted_ll, formatted_hl = row[f"formatted_{ll}"], row[f"formatted_{hl}"]
            formatted_ci = "".join([caps[0], formatted_ll, connector, formatted_hl, caps[1]])
            dataframe.loc[ix, "ci_range"] = formatted_ci
            dataframe.loc[ix, "est_ci"] = "".join([formatted_est, formatted_ci])
    return dataframe


def star_pval(
    dataframe: pd.core.frame.DataFrame,
    pval: str,
    starpval: bool,
    decimal_precision: int,
    **kwargs: Any,
) -> pd.core.frame.DataFrame:
    """Star the p-values according to the thresholds and symbols.

    Creating only on demand.

    Parameters
    ----------
    dataframe (pandas.core.frame.DataFrame)
            Pandas DataFrame where rows are variables. Columns are variable name, estimates,
            margin of error, etc.
    pval (str)
            Name of column containing the p-values.
    starpval (bool)
            If True, use 'thresholds' and 'symbols' to "star" the p-values.
    decimal_precision (int)
            Precision of 2 means we go from '0.1234' -> '0.12'.
    thresholds (list-like)
            List of thresholds to star p-values. Default, which is pretty conventional, is
            0.01, 0.05, 0.1.
    symbols (list-like)
            List of symbols corresponding to the stated thresholds. Can be stars (e.g. ***
            or can be letters, e.g. 'c', 'b', 'a'

    Returns
    -------
            pd.core.frame.DataFrame with additional column for 'formatted_pval'.
    """
    if pval is not None:
        thresholds = kwargs.get("thresholds", (0.01, 0.05, 0.1))
        symbols = kwargs.get("symbols", ("***", "**", "*"))

        check_iterables_samelen(thresholds, symbols)
        for ix, row in dataframe.iterrows():
            val = row[pval]
            if (np.isnan(val)) or (val == "nan") or (val == ""):
                dataframe.loc[ix, "formatted_pval"] = ""
                continue

            dec_formatted_pval = round(val, decimal_precision)
            dataframe.loc[ix, "formatted_pval"] = str(dec_formatted_pval)
            if starpval:
                for iy, threshold in enumerate(thresholds):
                    if val <= threshold:
                        _pval = "".join([str(dec_formatted_pval), symbols[iy]])
                        dataframe.loc[ix, "formatted_pval"] = _pval
                        break
    return dataframe


def indent_nongroupvar(
    dataframe: pd.core.frame.DataFrame,
    varlabel: str,
    groupvar: str,
    varindent: int = 2,
) -> pd.core.frame.DataFrame:
    """Indent the non-group variable labels when the 'form_ci_report' option is switched off.

    Creating only on demand.

    Parameters
    ----------
    dataframe (pandas.core.frame.DataFrame)
            Pandas DataFrame where rows are variables. Columns are variable name, estimates,
            margin of error, etc.
    varlabel (str)
            Name of column containing the variable label to be printed out.
    groupvar (str)
            Name of column containing group of variables.
    varindent (int)
            Amount of whitespace to indent the variables when grouping of variables is used.

    Returns
    -------
            pd.core.frame.DataFrame with nongroup variables in 'varlabel' indented by stated amount.
    """
    if varindent > 0:
        if groupvar is not None:
            groups = [gr.lower() for gr in dataframe[groupvar].unique()]

        for ix, row in dataframe.iterrows():
            label = row[varlabel]
            if (groupvar is not None) and (label.lower() not in groups):
                label = label.rjust(varindent + len(label))
            else:
                label = label.rjust(len(label))
            dataframe.loc[ix, varlabel] = label
    return dataframe


def normalize_varlabels(
    dataframe: pd.core.frame.DataFrame,
    varlabel: str,
    capitalize: str = "capitalize",
) -> pd.core.frame.DataFrame:
    """
    Normalize variable labels to capitalize or title form.

    Parameters
    ----------
    dataframe (pandas.core.frame.DataFrame)
            Pandas DataFrame where rows are variables. Columns are variable name, estimates,
            margin of error, etc.
    varlabel (str)
            Name of column containing the variable label to be printed out.
    capitalize (str)
            'capitalize' or 'title'
            See https://pandas.pydata.org/docs/reference/api/pandas.Series.str.capitalize.html

    Returns
    -------
            pd.core.frame.DataFrame with the varlabel column normalized.
    """
    if capitalize:
        if capitalize == "title":
            dataframe[varlabel] = dataframe[varlabel].str.title()
        elif capitalize == "capitalize":
            dataframe[varlabel] = dataframe[varlabel].str.capitalize()
        elif capitalize == "lower":
            dataframe[varlabel] = dataframe[varlabel].str.lower()
        elif capitalize == "upper":
            dataframe[varlabel] = dataframe[varlabel].str.upper()
        elif capitalize == "swapcase":
            dataframe[varlabel] = dataframe[varlabel].str.swapcase()
    return dataframe


def format_varlabels(
    dataframe: pd.core.frame.DataFrame,
    varlabel: str,
    form_ci_report: bool,
    ci_report: bool,
    groupvar: Union[str, None],
    extrapad: int = 2,
) -> pd.core.frame.DataFrame:
    """Format the yticklabels as normalized variable labels + estimate + confidence interval if form_ci_report = True. If form_ci_report = False, then just use the variable label.

    Creating only on demand.

    Parameters
    ----------
    dataframe (pandas.core.frame.DataFrame)
            Pandas DataFrame where rows are variables. Columns are variable name, estimates,
            margin of error, etc.
    varlabel (str)
            Name of column containing the variable label to be printed out.
    form_ci_report (bool)
            If True, form the formatted confidence interval as a string.
    ci_report (bool)
            If True, form the formatted confidence interval as a string.
    column2 (str)
            Name of column containing the second column of annotation to be printed.
    groupvar (str)
            Name of column containing group of variables.
    extrapad (int)
            Amount of padding between variable label and the estimate + confidence interval formatted string.
    varindent (int)
            Amount of whitespace to indent the variables when grouping of variables is used.

    Helpers
    -------
            _get_max_varlen
            _format_secondcolumn
            _remove_est_ci

    Returns
    -------
            pd.core.frame.DataFrame with an additional 'yticklabel' column.
    """
    if form_ci_report:
        pad = _get_max_varlen(dataframe=dataframe, varlabel=varlabel, extrapad=extrapad)
        for ix, row in dataframe.iterrows():
            var = row[varlabel]
            if ci_report:
                est_ci = row["est_ci"]
            else:
                est_ci = ""
                pad = 0

            if groupvar is not None:
                groups = [gr.lower() for gr in dataframe[groupvar].unique()]
                if var.lower() in groups:  # If row is a group header
                    yticklabel = var
                else:  # indent variable labels
                    yticklabel = "".join([var.ljust(pad), est_ci])
            else:
                yticklabel = "".join([var.ljust(pad), est_ci])
            dataframe.loc[ix, "yticklabel"] = yticklabel
    else:
        dataframe["yticklabel"] = dataframe[varlabel]
    dataframe = _remove_est_ci(dataframe=dataframe, varlabel=varlabel, groupvar=groupvar)
    return dataframe


def _remove_est_ci(
    dataframe: pd.core.frame.DataFrame,
    varlabel: str,
    groupvar: Optional[str],
) -> pd.core.frame.DataFrame:
    """
    Make rows for 'est_ci' and 'ci_range' empty string '' if row is a group variable label.

    Parameters
    ----------
    dataframe (pandas.core.frame.DataFrame)
            Pandas DataFrame where rows are variables. Columns are variable name, estimates,
            margin of error, etc.
    varlabel (str)
            Name of column containing the variable label to be printed out.
    groupvar (str)
            Name of column containing group of variables.

    Returns
    -------
            pd.core.frame.DataFrame.
    """
    if groupvar is not None:
        for ix, row in dataframe.iterrows():
            var = row[varlabel]
            grouplabel = row[groupvar]
            if var.lower().strip() == grouplabel.lower().strip():  # If row is a group header
                dataframe.loc[ix, "ci_range"], dataframe.loc[ix, "est_ci"] = "", ""
    return dataframe


def _get_max_varlen(
    dataframe: pd.core.frame.DataFrame,
    varlabel: str,
    extrapad: int,
) -> int:
    """
    Get max variable length in dataframe.

    Parameters
    ----------
    dataframe (pandas.core.frame.DataFrame)
            Pandas DataFrame where rows are variables. Columns are variable name, estimates,
            margin of error, etc.
    varlabel (str)
            Name of column containing the variable label to be printed out.
    extrapad (int)
            Amount of padding between variable label and the estimate + confidence interval formatted string.

    Returns
    -------
            int
    """
    max_varlen = dataframe[varlabel].map(str).str.len().max()
    return max_varlen + extrapad


def prep_annote(
    dataframe: pd.core.frame.DataFrame,
    annote: Optional[Union[Sequence[str], None]],
    annoteheaders: Optional[Union[Sequence[str], None]],
    varlabel: str,
    groupvar: str,
    **kwargs: Any,
) -> pd.core.frame.DataFrame:
    """Prepare the additional columns to be printed as annotations.

    Creating only on demand.

    Parameters
    ----------
    dataframe (pandas.core.frame.DataFrame)
            Pandas DataFrame where rows are variables. Columns are variable name, estimates,
            margin of error, etc.
    annote (list-like)
            List of columns to add as additional annotation in the plot.
    annoteheaders (list-like)
            List of table headers to use as column headers for the additional annotations.
    varlabel (str)
            Name of column containing the variable label to be printed out.
    groupvar (str)
            Name of column containing group of variables.

    Helpers
    -------
            _get_max_varlen

    Returns
    -------
            pd.core.frame.DataFrame with an additional formatted 'yticklabel' column.
    """
    col_spacing = kwargs.get("col_spacing", 2)

    lookup_annote_len = {}
    for ix, annotation in enumerate(annote):
        # Get max len for padding
        _pad = _get_max_varlen(dataframe=dataframe, varlabel=annotation, extrapad=0)
        if annoteheaders is not None:  # Check that max len exceeds header length
            _header = annoteheaders[ix]
            _pad = max(_pad, len(_header))
        lookup_annote_len[ix] = _pad
        for iy, row in dataframe.iterrows():  # Make individual formatted_annotations
            _annotation = str(row[annotation]).ljust(_pad)
            dataframe.loc[iy, f"formatted_{annotation}"] = _annotation

    # get max length for variables
    pad = _get_max_varlen(dataframe=dataframe, varlabel=varlabel, extrapad=0)

    if groupvar is not None:
        groups = [gr.lower() for gr in dataframe[groupvar].unique()]
    else:
        groups = []

    for ix, row in dataframe.iterrows():
        yticklabel = row[varlabel]
        if yticklabel.lower().strip() in groups:
            dataframe.loc[ix, "yticklabel"] = yticklabel
        else:
            yticklabel = yticklabel.ljust(pad)
            for annotation in annote:
                _annotation = row[f"formatted_{annotation}"]
                spacing = "".ljust(col_spacing)
                yticklabel = spacing.join([yticklabel, _annotation])
            dataframe.loc[ix, "yticklabel"] = yticklabel
    return dataframe


def prep_rightannnote(
    dataframe: pd.core.frame.DataFrame,
    rightannote: Optional[Union[Sequence[str], None]],
    right_annoteheaders: Optional[Union[Sequence[str], None]],
    varlabel: str,
    groupvar: str,
    **kwargs: Any,
) -> pd.core.frame.DataFrame:
    """Prepare the additional columns to be printed as annotations on the right.

    Creating only on demand.

    Parameters
    ----------
    dataframe (pandas.core.frame.DataFrame)
            Pandas DataFrame where rows are variables. Columns are variable name, estimates,
            margin of error, etc.
    rightannote (list-like)
            List of columns to add as additional annotation on the right-hand side of the plot.
    right_annoteheaders (list-like)
            List of table headers to use as column headers for the additional annotations.
    varlabel (str)
            Name of column containing the variable label to be printed out.
    groupvar (str)
            Name of column containing group of variables.

    Helpers
    -------
            _get_max_varlen

    Returns
    -------
            pd.core.frame.DataFrame with an additional formatted 'yticklabel2' column.
    """
    col_spacing = kwargs.get("col_spacing", 2)

    lookup_annote_len = {}
    for ix, annotation in enumerate(rightannote):
        # Get max len for padding
        _pad = _get_max_varlen(dataframe=dataframe, varlabel=annotation, extrapad=0)
        if right_annoteheaders is not None:  # Check that max len exceeds header length
            _header = right_annoteheaders[ix]
            _pad = max(_pad, len(_header))
        lookup_annote_len[ix] = _pad

        for iy, row in dataframe.iterrows():
            _annotation = row[annotation]
            _annotation = str(_annotation).ljust(_pad)
            dataframe.loc[iy, f"formatted_{annotation}"] = _annotation

    if groupvar is not None:
        groups = [gr.lower() for gr in dataframe[groupvar].unique()]
    else:
        groups = []

    for ix, row in dataframe.iterrows():
        yticklabel = row[varlabel]
        yticklabel2 = ""
        if yticklabel.lower().strip() in groups:
            dataframe.loc[ix, "yticklabel2"] = ""
        else:
            for annotation in rightannote:
                _annotation = row[f"formatted_{annotation}"]
                spacing = "".ljust(col_spacing)
                if yticklabel2 == "":
                    yticklabel2 = _annotation
                else:
                    yticklabel2 = spacing.join([yticklabel2, _annotation])
            dataframe.loc[ix, "yticklabel2"] = yticklabel2
    return dataframe


def make_tableheaders(
    dataframe: pd.core.frame.DataFrame,
    varlabel: str,
    annote: Optional[Union[Sequence[str], None]],
    annoteheaders: Optional[Union[Sequence[str], None]],
    rightannote: Optional[Union[Sequence[str], None]],
    right_annoteheaders: Optional[Union[Sequence[str], None]],
    **kwargs: Any,
) -> pd.core.frame.DataFrame:
    """Make the table headers from 'annoteheaders' and 'right_annoteheaders' as a row in the dataframe.

    Creating only on demand.

    Parameters
    ----------
    dataframe (pandas.core.frame.DataFrame)
            Pandas DataFrame where rows are variables. Columns are variable name, estimates,
            margin of error, etc.
    varlabel (str)
            Name of column containing the variable label to be printed out.
    annote (list-like)
            List of columns to add as additional annotation in the plot.
    annoteheaders (list-like)
            List of table headers to use as column headers for the additional annotations.
    rightannote (list-like)
            List of columns to add as additional annotation on the right-hand side of the plot.
    right_annoteheaders (list-like)
            List of table headers to use as column headers for the additional annotations.

    Helpers
    -------
            _get_max_varlen

    Returns
    -------
            pd.core.frame.DataFrame with an additional row as the table headers.
    """
    # No table headers
    variable_header = kwargs.get("variable_header", "")
    if (variable_header == "") or (variable_header is None):
        if (annoteheaders is None) and (right_annoteheaders is None):
            return dataframe

    # Make tableheaders
    col_spacing = kwargs.get("col_spacing", 2)
    spacing = "".ljust(col_spacing)

    variable_header = kwargs.get("variable_header", "Variable")
    dataframe = insert_empty_row(dataframe)

    pad = _get_max_varlen(dataframe=dataframe, varlabel=varlabel, extrapad=0)
    left_headers = variable_header.ljust(pad)
    dataframe.loc[0, "yticklabel"] = left_headers
    if annoteheaders is not None:
        for ix, header in enumerate(annoteheaders):
            corresponding_col = annote[ix]
            # get max length for variables
            pad = _get_max_varlen(dataframe=dataframe, varlabel=corresponding_col, extrapad=0)
            pad = max(pad, len(header))
            left_headers = spacing.join([left_headers, header.ljust(pad)])
        dataframe.loc[0, "yticklabel"] = left_headers

    if right_annoteheaders is not None:
        right_headers = ""
        for ix, header in enumerate(right_annoteheaders):
            corresponding_col = rightannote[ix]
            # get max length for variables
            pad = _get_max_varlen(dataframe=dataframe, varlabel=corresponding_col, extrapad=0)
            pad = max(pad, len(header))
            if right_headers == "":
                right_headers = header.ljust(pad)
            else:
                right_headers = spacing.join([right_headers, header.ljust(pad)])
        dataframe.loc[0, "yticklabel2"] = right_headers
    else:
        dataframe.loc[0, "yticklabel2"] = ""
    return dataframe


def _right_justify_num(
    dataframe: pd.core.frame.DataFrame, col: str, decimal_precision: int
) -> pd.core.frame.DataFrame:
    """
    Format numeric columns according to the amount of decimal precision and variable length.

    Parameters
    ----------
    dataframe (pandas.core.frame.DataFrame)
            Pandas DataFrame where rows are variables. Columns are variable name, estimates,
            margin of error, etc.
    col (str)
            Name of numeric column to format.
    decimal_precision (int)
            Precision of 2 means we go from '0.1234' -> '0.12'.

    Helpers
    -------
            _get_max_varlen

    Returns
    -------
            pd.core.frame.DataFrame with additional column for the formatted numeric column.
    """
    dataframe[f"formatted_{col}"] = (
        dataframe[col].apply(lambda x: f"{x:0.{decimal_precision}f}").astype(str)
    )
    pad = _get_max_varlen(dataframe=dataframe, varlabel=f"formatted_{col}", extrapad=0)
    dataframe[f"formatted_{col}"] = dataframe[f"formatted_{col}"].apply(lambda x: x.rjust(pad))
    return dataframe



"""Main functions to plot the forest plots."""
from typing import Any, List, Optional, Sequence, Tuple, Union

import numpy as np
import pandas as pd

np.seterr(all="ignore")
import matplotlib.pyplot as plt
from matplotlib import rcParams
from matplotlib.pyplot import Axes

from forestplot.arg_validators import check_data
from forestplot.dataframe_utils import insert_groups, reverse_dataframe, sort_data, sort_groups
from forestplot.graph_utils import (
    despineplot,
    draw_alt_row_colors,
    draw_ci,
    draw_est_markers,
    draw_pval_right,
    draw_ref_xline,
    draw_tablelines,
    draw_ylabel1,
    draw_yticklabel2,
    format_grouplabels,
    format_tableheader,
    format_xlabel,
    format_xticks,
    remove_ticks,
    right_flush_yticklabels,
)
from forestplot.text_utils import (
    form_est_ci,
    format_varlabels,
    indent_nongroupvar,
    make_tableheaders,
    normalize_varlabels,
    prep_annote,
    prep_rightannnote,
    star_pval,
)

rcParams["font.monospace"] = [
    "Lucida Sans Typewriter",
    "DejaVu Sans Mono",
    "Courier New",
    "Lucida Console",
]


def forestplot(
    dataframe: pd.core.frame.DataFrame,
    estimate: str,
    varlabel: str,
    ll: Optional[str] = None,
    hl: Optional[str] = None,
    form_ci_report: bool = True,
    ci_report: bool = True,
    groupvar: Optional[str] = None,
    group_order: Optional[Union[list, tuple]] = None,
    logscale: bool = False,
    annote: Optional[Union[Sequence[str], None]] = None,
    annoteheaders: Optional[Union[Sequence[str], None]] = None,
    rightannote: Optional[Union[Sequence[str], None]] = None,
    right_annoteheaders: Optional[Union[Sequence[str], None]] = None,
    pval: Optional[str] = None,
    capitalize: Optional[str] = None,
    starpval: bool = True,
    sort: bool = False,
    sortby: Optional[str] = None,
    flush: bool = True,
    decimal_precision: int = 2,
    figsize: Union[Tuple, List] = (4, 8),
    xticks: Optional[Union[list, range]] = None,
    ylabel: Optional[str] = None,
    xlabel: Optional[str] = None,
    yticker2: Optional[str] = None,
    color_alt_rows: bool = False,
    return_df: bool = False,
    preprocess: bool = True,
    table: bool = False,
    **kwargs: Any,
) -> Axes:
    """
    Draw forest plot using the pandas dataframe provided.

    Parameters
    ----------
    dataframe (pandas.core.frame.DataFrame)
            Pandas DataFrame where rows are variables. Columns are variable name, estimates,
            margin of error, etc.
    estimate (str)
            Name of column containing the estimates (e.g. pearson correlation coefficient,
            OR, regression estimates, etc.).
    varlabel (str)
            Name of column containing the variable label to be printed out.
    ll (str)
            Name of column containing the lower limit of the confidence intervals.
    hl (str)
            Name of column containing the upper limit of the confidence intervals.
    form_ci_report (bool)
            If True, form the formatted confidence interval as a string.
    ci_report (bool)
            If True, report the formatted confidence interval as a string.
    groupvar (str)
            Name of column containing group of variables.
    group_order (list-like)
            List of groups by order to report in the figure.
    logscale (bool)
            If True, plot the x-axis on log scale (base 10).
    annote (list-like)
            List of columns to add as additional annotation on the left-hand side of the plot.
    annoteheaders (list-like)
            List of table headers to use as column headers for the additional annotations
            on the left-hand side of the plot.
    rightannote (list-like)
            List of columns to add as additional annotation on the right-hand side of the plot.
    right_annoteheaders (list-like)
            List of table headers to use as column headers for the additional annotations
            on the right-hand side of the plot.
    pval (str)
            Name of column containing the p-values.
    starpval (bool)
            If True, use 'thresholds' and 'symbols' to "star" the p-values.
    sort (bool)
            If True, sort rows by estimate size
    sortby (str)
            Name of column to sort the dataframe by. Default is 'estimate'.
    flush (bool)
            Left-flush the variable labels.
    decimal_precision (int)
            Precision of 2 means we go from '0.1234' -> '0.12'.
    figsize (list-like):
            Figure size setting. E.g. (5,10) means width-to-height is 5 to 10.
            Size is for the dot-and-whisker plot region only. Does not control eventual
            figure size that comes from the length of the right and left y-axis ticker labels.
    xticks (list-like)
            List of xtickers to print on the x-axis.
    ylabel (str)
            Title of the left-hand side y-axis.
    xlabel (str)
            Title of the left-hand side x-axis.
    yticker2 (str)
            Name of column containing the second set of values to print on the right-hand side ytickers.
            If 'pval' is provided, then yticker2 will be set to the 'formatted_pval'.
    color_alt_rows (bool)
            If True, color alternative rows.
    preprocess (bool)
            If True, call the preprocess_dataframe() function to prepare the data for plotting.
    return_df (bool)
            If True, in addition to the Matplotlib Axes object, returns the intermediate dataframe
            created from preprocess_dataframe().
            A tuple of (preprocessed_dataframe, Ax) will be returned.

    Returns
    -------
            Matplotlib Axes object.
    """
    _local_df = dataframe.copy(deep=True)
    _local_df = check_data(
        dataframe=_local_df,
        estimate=estimate,
        varlabel=varlabel,
        pval=pval,
        ll=ll,
        hl=hl,
        groupvar=groupvar,
        group_order=group_order,
        annote=annote,
        annoteheaders=annoteheaders,
        rightannote=rightannote,
        right_annoteheaders=right_annoteheaders,
    )
    if ll is None:
        ci_report = False
    if ci_report is True:
        form_ci_report = True
    if preprocess:
        _local_df = _preprocess_dataframe(
            dataframe=_local_df,
            estimate=estimate,
            varlabel=varlabel,
            ll=ll,
            hl=hl,
            form_ci_report=form_ci_report,
            ci_report=ci_report,
            capitalize=capitalize,
            groupvar=groupvar,
            group_order=group_order,
            annote=annote,
            annoteheaders=annoteheaders,
            rightannote=rightannote,
            right_annoteheaders=right_annoteheaders,
            pval=pval,
            starpval=starpval,
            sort=sort,
            sortby=sortby,
            flush=flush,
            decimal_precision=decimal_precision,
            **kwargs,
        )
    ax = _make_forestplot(
        dataframe=_local_df,
        yticklabel="yticklabel",
        estimate=estimate,
        groupvar=groupvar,
        logscale=logscale,
        annoteheaders=annoteheaders,
        rightannote=rightannote,
        right_annoteheaders=right_annoteheaders,
        pval=pval,
        figsize=figsize,
        xticks=xticks,
        ll=ll,
        hl=hl,
        flush=flush,
        ylabel=ylabel,
        xlabel=xlabel,
        yticker2=yticker2,
        color_alt_rows=color_alt_rows,
        table=table,
        **kwargs,
    )
    return (_local_df, ax) if return_df else ax


def _preprocess_dataframe(
    dataframe: pd.core.frame.DataFrame,
    estimate: str,
    varlabel: str,
    ll: Optional[str] = None,
    hl: Optional[str] = None,
    form_ci_report: Optional[bool] = False,
    ci_report: Optional[bool] = False,
    groupvar: Optional[str] = None,
    group_order: Optional[Union[list, tuple]] = None,
    annote: Optional[Union[Sequence[str], None]] = None,
    annoteheaders: Optional[Union[Sequence[str], None]] = None,
    rightannote: Optional[Union[Sequence[str], None]] = None,
    right_annoteheaders: Optional[Union[Sequence[str], None]] = None,
    capitalize: Optional[str] = None,
    pval: Optional[str] = None,
    starpval: bool = True,
    sort: bool = False,
    sortby: Optional[str] = None,
    sortascend: bool = True,
    flush: bool = True,
    decimal_precision: int = 2,
    **kwargs: Any,
) -> pd.core.frame.DataFrame:
    """
    Preprocess the dataframe to be ready for plotting.

    Returns
    -------
            pd.core.frame.DataFrame with additional columns for plotting.
    """
    if (groupvar is not None) and (group_order is not None):
        if sort is True:
            list(group_order).reverse()
        dataframe = sort_groups(dataframe, groupvar=groupvar, group_order=group_order)
    dataframe = sort_data(
        dataframe=dataframe,
        estimate=estimate,
        groupvar=groupvar,
        sort=sort,
        sortby=sortby,
        sortascend=sortascend,
    )
    if groupvar is not None:  # Make groups
        dataframe = normalize_varlabels(
            dataframe=dataframe, varlabel=groupvar, capitalize=capitalize
        )
        dataframe = insert_groups(dataframe=dataframe, groupvar=groupvar, varlabel=varlabel)
    dataframe = normalize_varlabels(
        dataframe=dataframe, varlabel=varlabel, capitalize=capitalize
    )
    dataframe = indent_nongroupvar(dataframe=dataframe, varlabel=varlabel, groupvar=groupvar)
    if form_ci_report:
        dataframe = form_est_ci(
            dataframe=dataframe,
            estimate=estimate,
            ll=ll,
            hl=hl,
            decimal_precision=decimal_precision,
        )
    dataframe = star_pval(
        dataframe=dataframe,
        pval=pval,
        starpval=starpval,
        decimal_precision=decimal_precision,
        **kwargs,
    )
    if annote is None:  # Form ytickers = formatted variable labels
        dataframe = format_varlabels(
            dataframe=dataframe,
            varlabel=varlabel,
            form_ci_report=form_ci_report,
            ci_report=ci_report,
            groupvar=groupvar,
        )
    else:
        dataframe = prep_annote(
            dataframe=dataframe,
            groupvar=groupvar,
            varlabel=varlabel,
            annote=annote,
            annoteheaders=annoteheaders,
            **kwargs,
        )
    if rightannote is not None:
        dataframe = prep_rightannnote(
            dataframe=dataframe,
            groupvar=groupvar,
            varlabel=varlabel,
            rightannote=rightannote,
            right_annoteheaders=right_annoteheaders,
        )
    dataframe = make_tableheaders(
        dataframe=dataframe,
        varlabel=varlabel,
        annote=annote,
        annoteheaders=annoteheaders,
        rightannote=rightannote,
        right_annoteheaders=right_annoteheaders,
        **kwargs,
    )
    return reverse_dataframe(dataframe)  # since plotting starts from bottom


def _make_forestplot(
    dataframe: pd.core.frame.DataFrame,
    yticklabel: str,
    estimate: str,
    groupvar: str,
    pval: str,
    xticks: Optional[Union[list, range]],
    ll: str,
    hl: str,
    logscale: bool,
    flush: bool,
    annoteheaders: Optional[Union[Sequence[str], None]],
    rightannote: Optional[Union[Sequence[str], None]],
    right_annoteheaders: Optional[Union[Sequence[str], None]],
    ylabel: str,
    xlabel: str,
    yticker2: Optional[str],
    color_alt_rows: bool,
    figsize: Union[Tuple, List],
    despine: bool = True,
    table: bool = False,
    **kwargs: Any,
) -> Axes:
    """
    Draw the forest plot.

    Returns
    -------
            Matplotlib Axes object.
    """
    _, ax = plt.subplots(figsize=figsize, facecolor="white")
    ax = draw_ci(
        dataframe=dataframe,
        estimate=estimate,
        yticklabel=yticklabel,
        ll=ll,
        hl=hl,
        logscale=logscale,
        ax=ax,
        **kwargs,
    )
    draw_est_markers(
        dataframe=dataframe, estimate=estimate, yticklabel=yticklabel, ax=ax, **kwargs
    )
    format_xticks(
        dataframe=dataframe, estimate=estimate, ll=ll, hl=hl, xticks=xticks, ax=ax, **kwargs
    )
    draw_ref_xline(
        ax=ax,
        dataframe=dataframe,
        annoteheaders=annoteheaders,
        right_annoteheaders=right_annoteheaders,
        **kwargs,
    )
    pad = right_flush_yticklabels(
        dataframe=dataframe, yticklabel=yticklabel, flush=flush, ax=ax, **kwargs
    )
    if rightannote is None:
        ax, righttext_width = draw_pval_right(
            dataframe=dataframe,
            pval=pval,
            annoteheaders=annoteheaders,
            rightannote=rightannote,
            yticklabel=yticklabel,
            yticker2=yticker2,
            pad=pad,
            ax=ax,
            **kwargs,
        )
    else:
        ax, righttext_width = draw_yticklabel2(
            dataframe=dataframe,
            annoteheaders=annoteheaders,
            right_annoteheaders=right_annoteheaders,
            ax=ax,
            **kwargs,
        )

    draw_ylabel1(ylabel=ylabel, pad=pad, ax=ax, **kwargs)
    remove_ticks(ax)
    format_grouplabels(dataframe=dataframe, groupvar=groupvar, ax=ax, **kwargs)
    format_tableheader(
        annoteheaders=annoteheaders, right_annoteheaders=right_annoteheaders, ax=ax, **kwargs
    )
    despineplot(despine=despine, ax=ax)
    format_xlabel(xlabel=xlabel, ax=ax, **kwargs)
    if color_alt_rows:
        draw_alt_row_colors(
            dataframe,
            groupvar=groupvar,
            annoteheaders=annoteheaders,
            right_annoteheaders=right_annoteheaders,
            ax=ax,
        )
    if (annoteheaders is not None) or (
        (pval is not None) or (right_annoteheaders is not None)
    ):
        if table:
            draw_tablelines(
                dataframe=dataframe,
                righttext_width=righttext_width,
                pval=pval,
                right_annoteheaders=right_annoteheaders,
                ax=ax,
            )
    if annoteheaders or right_annoteheaders:
        negative_padding = 1.0
    else:
        negative_padding = 0.5
    ax.set_ylim(-0.5, ax.get_ylim()[1] - negative_padding)
    return ax



"""Holds functions to check prepare dataframe for plotting."""
from typing import Any, Optional, Union

import numpy as np
import pandas as pd


def insert_groups(
    dataframe: pd.core.frame.DataFrame, groupvar: str, varlabel: str
) -> pd.core.frame.DataFrame:
    """
    Insert the name of variable groups as a psuedo variable in the dataframe.

    Parameters
    ----------
    dataframe (pandas.core.frame.DataFrame)
            Pandas DataFrame where rows are variables. Columns are variable name, estimates,
            margin of error, etc.
    groupvar (str)
            Name of column containing group of variables.
    varlabel (str)
            Name of column containing the variable label to be printed out.

    Returns
    -------
            pd.core.frame.DataFrame with group variable labels inserted as psuedo variables.
    """
    df_groupsasvar = pd.DataFrame()
    for group in dataframe[groupvar].unique():
        _df = dataframe.query(f"{groupvar}==@group")
        addgroupvar = pd.DataFrame({varlabel: [group], groupvar: [group]})
        df_groupsasvar = pd.concat([df_groupsasvar, addgroupvar, _df], ignore_index=True)
    return df_groupsasvar


def sort_groups(
    dataframe: pd.core.frame.DataFrame, groupvar: str, group_order: Union[list, tuple]
) -> pd.core.frame.DataFrame:
    """
    Sort dataframe by list of groups implying order.

    Parameters
    ----------
    dataframe (pandas.core.frame.DataFrame)
            Pandas DataFrame where rows are variables. Columns are variable name, estimates,
            margin of error, etc.
    groupvar (str)
            Name of column containing group of variables.
    group_order (list-like)
            List of groups by order to report in the figure.

    Returns
    -------
            pd.core.frame.DataFrame	ordered by order in 'group_order'.
    """
    dataframe[groupvar] = pd.Categorical(dataframe[groupvar], group_order)
    dataframe.sort_values(groupvar, inplace=True)
    return dataframe


def sort_data(
    dataframe: pd.core.frame.DataFrame,
    estimate: str,
    groupvar: str,
    sort: bool = False,
    sortby: Optional[str] = None,
    sortascend: bool = True,
    **kwargs: Any,
) -> pd.core.frame.DataFrame:
    """
    Sort the dataframe according to the stated options.

    Parameters
    ----------
    dataframe (pandas.core.frame.DataFrame)
            Pandas DataFrame where rows are variables. Columns are variable name, estimates,
            margin of error, etc.
    estimate (str)
            Name of column containing the estimates (e.g. pearson correlation coefficient,
            OR, regression estimates, etc.).
    groupvar (str)
            Name of column containing group of variables.
    sort (bool)
            If True, sort rows by estimate size
    sortby (str)
            Name of column to sort the dataframe by. Default is 'estimate'.
    sortascend (bool)
            Sort in ascending order.

    Returns
    -------
            pd.core.frame.DataFrame that is sorted by the stated options.
    """
    if sort or (sortby is not None):
        sortascend = not sortascend  # reverse b/c plot stands from bottom of df
        if sortby is None:
            sortby = estimate

        if groupvar is not None:
            dataframe.sort_values(
                [groupvar, sortby], ascending=[True, sortascend], inplace=True
            )
        else:
            dataframe.sort_values(sortby, ascending=sortascend, inplace=True)
        return dataframe.reset_index(drop=True)
    else:
        return dataframe


def reverse_dataframe(dataframe: pd.core.frame.DataFrame) -> pd.core.frame.DataFrame:
    """Flip the dataframe so that last row is now first and so on."""
    dataframe = dataframe[::-1]
    return dataframe.reset_index(drop=True)


def insert_empty_row(dataframe: pd.core.frame.DataFrame) -> pd.core.frame.DataFrame:
    """Add an empty row to the top of the dataframe."""
    _df = pd.DataFrame([[np.nan] * len(dataframe.columns)], columns=dataframe.columns)
    dataframe = pd.concat([_df, dataframe], axis=0, ignore_index=True)
    return dataframe


def load_data(name: str, **param_dict: Optional[Any]) -> pd.core.frame.DataFrame:
    """
    Load example dataset for quickstart.

    Example data available now:
            - mortality

    The source of these data will be from: https://github.com/LSYS/forestplot/tree/main/examples/data.

    Parameters
    ----------
    name (str)
            Name of the example data set.

    Returns
    -------
    pd.core.frame.DataFrame.
    """
    available_data = ["mortality", "sleep", "sleep-untruncated"]
    name = name.lower().strip()
    if name in available_data:
        url = (
            f"https://raw.githubusercontent.com/lsys/forestplot/main/examples/data/{name}.csv"
        )
        df = pd.read_csv(url, **param_dict)
        if name == "sleep":
            df["n"] = df["n"].astype("str")
        return df
    else:
        available_data_str = ", ".join(available_data)
        raise AssertionError(f"{name} not found. Should be one of '{available_data_str}'")



"""Holds functions to check data and validate arguments from users."""
import warnings
from typing import Optional, Sequence, Union

import pandas as pd
import pandas.api.types as ptypes


def check_data(
    dataframe: pd.core.frame.DataFrame,
    estimate: str,
    varlabel: str,
    groupvar: Optional[str] = None,
    group_order: Optional[Sequence] = None,
    ll: Optional[str] = None,
    hl: Optional[str] = None,
    form_ci_report: bool = None,
    annote: Optional[Union[Sequence[str], None]] = None,
    annoteheaders: Optional[Union[Sequence[str], None]] = None,
    rightannote: Optional[Union[Sequence[str], None]] = None,
    right_annoteheaders: Optional[Union[Sequence[str], None]] = None,
    pval: Optional[str] = None,
    ylabel2: Optional[str] = None,
) -> pd.core.frame.DataFrame:
    """Check and validate that dataframe has the correct data. If data is missing, create them.

    Checks and validates key arguments.

    Parameters
    ----------
    dataframe (pandas.core.frame.DataFrame)
            Pandas DataFrame where rows are variables. Columns are variable name, estimates,
            margin of error, etc.
    estimate (str)
            Name of column containing the estimates (e.g. pearson correlation coefficient,
            OR, regression estimates, etc.).
    varlabel (str)
            Name of column containing the variable label to be printed out.
    groupvar (str)
            Name of column containing group of variables.
    group_order (list-like)
            List of groups by order to report in the figure.
    ll (str)
            Name of column containing the lower limit of the confidence intervals.
    hl (str)
            Name of column containing the upper limit of the confidence intervals.
    form_ci_report (bool)
            If True, form the formatted confidence interval as a string.
    annote (list-like)
            List of columns to add as additional annotation in the plot.
    annoteheaders (list-like)
            List of table headers to use as column headers for the additional annotations.
    rightannote (list-like)
            List of columns to add as additional annotation on the right-hand side of the plot.
    right_annoteheaders (list-like)
            List of table headers to use as column headers for the additional annotations
            on the right-hand side of the plot.
    pval (str)
            Name of column containing the p-values.
    ylabel2 (str)
            Title of the right-hand side y-axis.

    Returns
    -------
            pd.core.frame.DataFrame.
    """
    ##########################################################################
    ## Check that CI options (ll, hl, form_ci_report) are consistent
    ##########################################################################
    if ll is None:
        try:
            assert hl is None
        except Exception:
            raise TypeError("'ll' is None. 'hl' should also be None.")

    if hl is None:
        try:
            assert ll is None
        except Exception:
            raise TypeError("'hl' is None. 'll' should also be None.")

    if ll is None and form_ci_report:
        warnings.warn("'ll' is None. 'form_ci_report' will be set to False.")

    ##########################################################################
    ## Check that numeric data are numeric
    ##########################################################################
    if not isinstance(dataframe, pd.core.frame.DataFrame):
        raise TypeError("Expect data as Pandas DataFrame")

    if not ptypes.is_numeric_dtype(dataframe[estimate]):
        try:
            dataframe[estimate] = dataframe[estimate].astype(float)
        except ValueError:
            raise TypeError("Estimates should be float or int")

    if (ll is not None) and (not ptypes.is_numeric_dtype(dataframe[ll])):
        try:
            dataframe[ll] = dataframe[ll].astype(float)
        except ValueError:
            raise TypeError("CI lowerlimit values should be float or int")

    if (hl is not None) and (not ptypes.is_numeric_dtype(dataframe[hl])):
        try:
            dataframe[hl] = dataframe[hl].astype(float)
        except ValueError:
            raise TypeError("CI higherlimit values should be float or int")

    ##########################################################################
    ## Check that the annotations and headers specified are list-like
    ##########################################################################
    if annote is not None:
        try:
            assert ptypes.is_list_like(annote)
        except Exception:
            raise TypeError("annote should be list-like.")

    if annoteheaders is not None:
        try:
            assert ptypes.is_list_like(annoteheaders)
        except Exception:
            raise TypeError("annoteheaders should be list-like.")

    if rightannote is not None:
        try:
            assert ptypes.is_list_like(rightannote)
        except Exception:
            raise TypeError("rightannote should be list-like.")

    if right_annoteheaders is not None:
        try:
            assert ptypes.is_list_like(right_annoteheaders)
        except Exception:
            raise TypeError("right_annoteheaders should be list-like.")

    ##########################################################################
    ## Check that annotations and corresponding headers have same length
    ##########################################################################
    if (annote is not None) & (annoteheaders is not None):
        check_iterables_samelen(annote, annoteheaders)

    if (rightannote is not None) & (right_annoteheaders is not None):
        check_iterables_samelen(rightannote, right_annoteheaders)

    ##########################################################################
    ## Check that specified annotations can be found in input or processed dataframe
    ##########################################################################
    acceptable_annotations = [  # from processed data
        "ci_range",
        "est_ci",
        "formatted_pval",
    ]

    if annote is not None:
        for col in annote:
            try:
                assert (col in dataframe.columns) or (col in acceptable_annotations)
            except Exception:
                raise AssertionError(f"the field {col} is not found in dataframe.")

    if rightannote is not None:
        for col in rightannote:
            try:
                assert (col in dataframe.columns) or (col in acceptable_annotations)
            except Exception:
                raise AssertionError(f"the field {col} is not found in dataframe.")

    if groupvar is not None:
        check_groups(dataframe, groupvar=groupvar, group_order=group_order)

    ##########################################################################
    ## Check that column names to add as annotations are provided if the
    ##   annotation headers are provided
    ##########################################################################
    if (annoteheaders is not None) and (annote is None):
        raise TypeError("Annotation headers are provided but no columns provided ('annote').")

    if (right_annoteheaders is not None) and (rightannote is None):
        raise TypeError(
            "Right annotation headers are provided but no columns provided ('rightannote')."
        )

    ##########################################################################
    ## Warnings
    ##########################################################################
    # Warn: Check that var itself is not in annote
    if (annote is not None) and (varlabel in annote):
        warnings.warn(
            f'{varlabel} is a variable is already printed. Specifying {varlabel} in "annote" will lead to duplicate printing of {varlabel}.'
        )

    if (rightannote is not None) and (varlabel in rightannote):
        warnings.warn(f"{varlabel} is a variable is already printed.")
        # warnings.warn(f'Specifying {varlabel} in "rightannote" will lead to duplicate printing of {varlabel}.')
        warnings.warn(
            f'{varlabel} is a variable is already printed. Specifying {varlabel} in "rightannote" will lead to duplicate printing of {varlabel}.'
        )

    if (annote is not None) and (rightannote is not None):
        if any(col in annote for col in rightannote):
            warnings.warn("Duplicates found in annote and rightannote.")

    # Overriding default to plot p-value on the right-hand side since rightannote is specified
    if (pval is not None) and (rightannote is not None):
        warnings.warn(
            "p-value will not be plotted in the right annotation column by default since rightannote is specificied."
        )

    # Warn: need to ignore ylabel2 if right_annote headers are specified
    if (ylabel2 is not None) and (right_annoteheaders is not None):
        warnings.warn("ylabel2 is ignored since right_annoteheaders is specified.")

    # Warn: duplicates found in varlabels and grouplabels
    if groupvar is not None:
        grouplabels = dataframe[groupvar].dropna().unique()
        grouplabels = [grplab_str.lower().strip() for grplab_str in grouplabels]
        varlabels = dataframe[varlabel].dropna().unique()
        varlabels = [varlab_str.lower().strip() for varlab_str in varlabels]
        if any(varlab_str in grouplabels for varlab_str in varlabels):
            warnings.warn(
                "Duplicates found in variable labels ('varlabel') and group labels ('groupvar'). Formatting of y-axis labels may lead to unexpected errors."
            )

    return dataframe


def check_iterables_samelen(*args):  # type: ignore
    """Assert that provided iterables have same length."""
    try:
        assert all(len(args[0]) == len(_arg) for _arg in args[1:])
    except Exception:
        raise ValueError("Iterables not of the same length.")
    return None


def check_groups(
    dataframe: pd.core.frame.DataFrame, groupvar: str, group_order: Optional[Sequence]
) -> None:
    """Check that provided group-related args are well-behaved.

    Parameters
    ----------
    dataframe (pandas.core.frame.DataFrame)
        Pandas DataFrame where rows are variables. Columns are variable name, estimates,
        margin of error, etc.
    groupvar (str)
        Name of column containing group of variables.
    group_order (list-like)
        List of groups by order to report in the figure.

    Returns
    -------
    None
    """
    if (group_order is not None) and (groupvar is None):
        raise TypeError(
            "Group ordering ('group_order') provided but no group column provided ('groupvar')."
        )
    # Check detected unique groups and group_order have same length
    groups: Sequence = dataframe[groupvar].dropna().unique()
    if group_order is not None:
        check_iterables_samelen(groups, group_order)
    # Check that groups in group_order exists
    if (group_order is not None) and (groupvar is not None):
        try:
            assert all(group in groups for group in group_order)
        except Exception:
            raise AssertionError("Groups specified in `group_order` should exist in the data.")
    return None



"""Holds functions to draw the plot."""
import warnings
from typing import Any, Optional, Sequence, Tuple, Union

import matplotlib.pyplot as plt
import pandas as pd
from matplotlib.pyplot import Axes

warnings.filterwarnings("ignore")


def draw_ci(
    dataframe: pd.core.frame.DataFrame,
    estimate: str,
    yticklabel: str,
    ll: str,
    hl: str,
    logscale: bool,
    ax: Axes,
    **kwargs: Any
) -> Axes:
    """
    Draw the confidence intervals using the Matplotlib errorbar API.

    Parameters
    ----------
    dataframe (pandas.core.frame.DataFrame)
            Pandas DataFrame where rows are variables. Columns are variable name, estimates,
            margin of error, etc.
    estimate (str)
            Name of column containing the estimates (e.g. pearson correlation coefficient,
            OR, regression estimates, etc.).
    yticklabel (str)
            Name of column in intermediate dataframe containing the formatted yticklabels.
    ll (str)
            Name of column containing the lower limit of the confidence intervals.
    hl (str)
            Name of column containing the upper limit of the confidence intervals.
    ax (Matplotlib Axes)
            Axes to operate on.

    Returns
    -------
            Matplotlib Axes object.
    """
    if ll is not None:
        lw = kwargs.get("lw", 1.4)
        linecolor = kwargs.get("linecolor", ".6")
        ax.errorbar(
            x=dataframe[estimate],
            y=dataframe[yticklabel],
            xerr=[dataframe[estimate] - dataframe[ll], dataframe[hl] - dataframe[estimate]],
            ecolor=linecolor,
            elinewidth=lw,
            ls="none",
            zorder=0,
        )
    if logscale:
        ax.set_xscale("log", base=10)
    return ax


def draw_est_markers(
    dataframe: pd.core.frame.DataFrame, estimate: str, yticklabel: str, ax: Axes, **kwargs: Any
) -> Axes:
    """
    Draws the markers of the estimates using the Matplotlib plt.scatter API.

    Parameters
    ----------
    dataframe (pandas.core.frame.DataFrame)
            Pandas DataFrame where rows are variables. Columns are variable name, estimates,
            margin of error, etc.
    estimate (str)
            Name of column containing the estimates (e.g. pearson correlation coefficient,
            OR, regression estimates, etc.).
    yticklabel (str)
            Name of column in intermediate dataframe containing the formatted yticklabels.
    ax (Matplotlib Axes)
            Axes to operate on.

    Returns
    -------
            Matplotlib Axes object.
    """
    marker = kwargs.get("marker", "s")
    markersize = kwargs.get("markersize", 40)
    markercolor = kwargs.get("markercolor", "darkslategray")
    markeralpha = kwargs.get("markeralpha", 0.8)
    ax.scatter(
        y=yticklabel,
        x=estimate,
        data=dataframe,
        marker=marker,
        s=markersize,
        color=markercolor,
        alpha=markeralpha,
    )
    return ax


def draw_ref_xline(
    ax: Axes,
    dataframe: pd.core.frame.DataFrame,
    annoteheaders: Optional[Union[Sequence[str], None]],
    right_annoteheaders: Optional[Union[Sequence[str], None]],
    **kwargs: Any
) -> Axes:
    """
    Draw the vertical reference xline at zero. Unless defaults are overridden in kwargs.

    Parameters
    ----------
    ax (Matplotlib Axes)
            Axes to operate on.

    Returns
    -------
            Matplotlib Axes object.
    """
    xline = kwargs.get("xline", 0)
    if xline is not None:
        xlinestyle = kwargs.get("xlinestyle", "-")
        xlinecolor = kwargs.get("xlinecolor", ".2")
        xlinewidth = kwargs.get("xlinewidth", 1)
        if (annoteheaders is None) and (right_annoteheaders is None):
            _offset = 0.5
        else:
            _offset = 1.5
        ax.vlines(
            x=xline,
            ymin=-0.5,
            ymax=len(dataframe) - _offset,
            linestyle=xlinestyle,
            color=xlinecolor,
            linewidth=xlinewidth,
        )
    return ax


def right_flush_yticklabels(
    dataframe: pd.core.frame.DataFrame, yticklabel: str, flush: bool, ax: Axes, **kwargs: Any
) -> float:
    """Flushes the formatted ytickers to the left. Also returns the amount of max padding in the window width.

    Padding to be used for drawing the 2nd yticklabels and ylabels.
    My reference: https://stackoverflow.com/questions/15882249/matplotlib-aligning-y-ticks-to-the-left

    Parameters
    ----------
    dataframe (pandas.core.frame.DataFrame)
            Pandas DataFrame where rows are variables. Columns are variable name, estimates,
            margin of error, etc.
    yticklabel (str)
            Name of column in intermediate dataframe containing the formatted yticklabels.
    flush (bool)
            Left-flush the variable labels.
    ax (Matplotlib Axes)
            Axes to operate on.

    Returns
    -------
            Window wdith of figure (float)
    """
    fontfamily = kwargs.get("fontfamily", "monospace")
    fontsize = kwargs.get("fontsize", 12)
    # plt.draw()
    fig = plt.gcf()
    if flush:
        ax.set_yticklabels(
            dataframe[yticklabel], fontfamily=fontfamily, fontsize=fontsize, ha="left"
        )
    else:
        ax.set_yticklabels(
            dataframe[yticklabel], fontfamily=fontfamily, fontsize=fontsize, ha="right"
        )
    yax = ax.get_yaxis()
    pad = max(
        T.label.get_window_extent(renderer=fig.canvas.get_renderer()).width
        for T in yax.majorTicks
    )
    if flush:
        yax.set_tick_params(pad=pad)

    return pad


def draw_pval_right(
    dataframe: pd.core.frame.DataFrame,
    pval: str,
    annoteheaders: Optional[Union[Sequence[str], None]],
    yticklabel: str,
    pad: float,
    ax: Axes,
    **kwargs: Any
) -> Tuple[Axes, float]:
    """
    Draws the 2nd ytick labels on the right-hand side of the figure.

    Parameters
    ----------
    dataframe (pandas.core.frame.DataFrame)
            Pandas DataFrame where rows are variables. Columns are variable name, estimates,
            margin of error, etc.
    pval (str)
            Name of column containing the p-values.
    annoteheaders (list-like)
            List of table headers to use as column headers for the additional annotations.
    yticklabel (str)
            Name of column in intermediate dataframe containing the formatted yticklabels.
    pad (float)
            Window wdith of figure
    ax (Matplotlib Axes)
            Axes to operate on.

    Returns
    -------
            Matplotlib Axes object.
    """
    if pval is not None:
        inv = ax.transData.inverted()
        righttext_width = 0
        fig = plt.gcf()
        for _, row in dataframe.iterrows():
            yticklabel1 = row[yticklabel]
            yticklabel2 = row["formatted_pval"]
            if pd.isna(yticklabel2):
                yticklabel2 = ""

            extrapad = 0.05
            pad = ax.get_xlim()[1] * (1 + extrapad)
            t = ax.text(
                x=pad,
                y=yticklabel1,
                s=yticklabel2,
                horizontalalignment="left",
                verticalalignment="center",
            )
            (_, _), (x1, _) = inv.transform(
                t.get_window_extent(renderer=fig.canvas.get_renderer())
            )
            righttext_width = max(righttext_width, x1)

        # 2nd label title
        pval_title = kwargs.get("pval_title", "P-value")

        # The next few lines of code make sure that the "P-value" label title on the right columns are of the
        # same height (using negative_padding) and fontsize (using kwargs.get("fontsize", 12)) as the
        # ylabel height and fontsize. See draw_ylabel1(...).
        fontsize = kwargs.get("fontsize", 12)
        ylabel1_size = kwargs.get("ylabel1_size", 1 + fontsize)
        if annoteheaders:
            negative_padding = 1.0
        else:
            negative_padding = 0.5

        if pval_title is not None:
            if annoteheaders is None:
                t = ax.text(
                    pad,
                    ax.get_ylim()[1] - negative_padding,
                    pval_title,
                    size=ylabel1_size,
                    fontweight="bold",
                )
                (_, _), (x1, _) = inv.transform(
                    t.get_window_extent(renderer=fig.canvas.get_renderer())
                )
                righttext_width = max(righttext_width, x1)
            if annoteheaders is not None:  # if tableheaders exist
                pval_title_fontweight = kwargs.get("pval_title_fontweight", "bold")
                pval_title_fontsize = kwargs.get("pval_title_fontsize", 12)

                header_index = len(ax.get_yticklabels()) - 1
                t = ax.text(
                    x=pad,
                    y=header_index,
                    s=pval_title,
                    size=pval_title_fontsize,
                    fontweight=pval_title_fontweight,
                    horizontalalignment="left",
                    verticalalignment="center",
                )
                (_, _), (x1, _) = inv.transform(
                    t.get_window_extent(renderer=fig.canvas.get_renderer())
                )
                righttext_width = max(righttext_width, x1)
        return ax, righttext_width
    else:
        return ax, 0


def draw_yticklabel2(
    dataframe: pd.core.frame.DataFrame,
    annoteheaders: Union[Sequence[str], None],
    right_annoteheaders: Union[Sequence[str], None],
    ax: Axes,
    **kwargs: Any
) -> Tuple[Axes, float]:
    """
    Draw the second ylabel title on the right-hand side y-axis.

    Parameters
    ----------
    dataframe (pandas.core.frame.DataFrame)
            Pandas DataFrame where rows are variables. Columns are variable name, estimates,
            margin of error, etc.
    ax (Matplotlib Axes)
            Axes to operate on.

    Returns
    -------
            Matplotlib Axes object.
    """
    grouplab_fontweight = kwargs.get("grouplab_fontweight", "bold")
    fontfamily = kwargs.get("fontfamily", "monospace")
    fontsize = kwargs.get("fontsize", 12)

    top_row_ix = len(dataframe) - 1
    inv = ax.transData.inverted()
    righttext_width = 0
    fig = plt.gcf()
    for ix, row in dataframe.iterrows():
        yticklabel1 = row["yticklabel"]
        yticklabel2 = row["yticklabel2"]

        extrapad = 0.05
        pad = ax.get_xlim()[1] * (1 + extrapad)
        if (ix == top_row_ix) and (
            annoteheaders is not None or right_annoteheaders is not None
        ):
            t = ax.text(
                x=pad,
                y=yticklabel1,
                s=yticklabel2,
                fontfamily=fontfamily,
                horizontalalignment="left",
                verticalalignment="center",
                fontsize=fontsize,
                fontweight=grouplab_fontweight,
            )
        else:
            t = ax.text(
                x=pad,
                y=yticklabel1,
                s=yticklabel2,
                fontfamily=fontfamily,
                horizontalalignment="left",
                verticalalignment="center",
                fontsize=fontsize,
            )
        (_, _), (x1, _) = inv.transform(
            t.get_window_extent(renderer=fig.canvas.get_renderer())
        )
        righttext_width = max(righttext_width, x1)
    return ax, righttext_width


def draw_ylabel1(ylabel: str, pad: float, ax: Axes, **kwargs: Any) -> Axes:
    """
    Draw ylabel title for the left-hand side y-axis.

    Parameters
    ----------
    ylabel (str)
            Title of the left-hand side y-axis.
    pad (float)
            Window wdith of figure
    ax (Matplotlib Axes)
            Axes to operate on.

    Returns
    -------
            Matplotlib Axes object.
    """
    fontsize = kwargs.get("fontsize", 12)
    ax.set_ylabel("")
    if ylabel is not None:
        # Retrieve settings from kwargs
        ylabel1_size = kwargs.get("ylabel1_size", 1 + fontsize)
        ylabel1_fontweight = kwargs.get("ylabel1_fontweight", "bold")
        ylabel_loc = kwargs.get("ylabel_loc", "top")
        ylabel_angle = kwargs.get("ylabel_angle", "horizontal")
        ax.set_ylabel(
            ylabel,
            loc=ylabel_loc,
            labelpad=-pad,
            rotation=ylabel_angle,
            size=ylabel1_size,
            fontweight=ylabel1_fontweight,
        )
    return ax


def remove_ticks(ax: Axes) -> Axes:
    """
    Remove the tickers on the top, left, and right borders.

    Parameters
    ----------
    ax (Matplotlib Axes)
            Axes to operate on.

    Returns
    -------
            Matplotlib Axes object.
    """
    ax.tick_params(
        top=False,
        bottom=True,
        left=False,
        right=False,
        labelleft=True,
        labelright=False,
        labelbottom=True,
    )
    return ax


def format_grouplabels(
    dataframe: pd.core.frame.DataFrame, groupvar: str, ax: Axes, **kwargs: Any
) -> Axes:
    """
    Bold the group variable labels.

    Fontweight options in Matplotlib: [ 'normal' | 'bold' | 'heavy' | 'light' | 'ultrabold' | 'ultralight' ]

    Parameters
    ----------
    dataframe (pandas.core.frame.DataFrame)
            Pandas DataFrame where rows are variables. Columns are variable name, estimates,
            margin of error, etc.
    groupvar (str)
            Name of column containing group of variables.
    ax (Matplotlib Axes)
            Axes to operate on.

    Returns
    -------
            Matplotlib Axes object.
    """
    grouplab_size = kwargs.get("grouplab_size", 12)
    grouplab_fontweight = kwargs.get("grouplab_fontweight", "bold")
    if groupvar is not None:
        for ix, ylabel in enumerate(ax.get_yticklabels()):
            for gr in dataframe[groupvar].unique():
                try:
                    if gr.lower() == ylabel.get_text().lower().strip():
                        ax.get_yticklabels()[ix].set_fontweight(grouplab_fontweight)
                        ax.get_yticklabels()[ix].set_fontsize(grouplab_size)
                        ax.get_yticklabels()[ix].set_fontfamily("sans-serif")
                except AttributeError:
                    pass
    return ax


def despineplot(despine: bool, ax: Axes) -> Axes:
    """
    Despine the plot by removing the top, left, and right borders.

    Parameters
    ----------
    despine (bool)
            If True, despine.
    ax (Matplotlib Axes)
            Axes to operate on.

    Returns
    -------
            Matplotlib Axes object.
    """
    if despine:
        ax.spines["top"].set_color("None")
        ax.spines["left"].set_color("None")
        ax.spines["right"].set_color("None")
    return ax


def format_tableheader(
    annoteheaders: Optional[Union[Sequence[str], None]],
    right_annoteheaders: Optional[Union[Sequence[str], None]],
    ax: Axes,
    **kwargs: Any
) -> Axes:
    """
    Format the tableheader as the first row in the data.

    Parameters
    ----------
    annoteheaders (list-like)
            List of table headers to use as column headers for the additional annotations
            on the left-hand side of the plot.
    right_annoteheaders (list-like)
            List of table headers to use as column headers for the additional annotations
            on the right-hand side of the plot.
    ax (Matplotlib Axes)
            Axes to operate on.

    Returns
    -------
            Matplotlib Axes object.
    """
    if (annoteheaders is not None) or (right_annoteheaders is not None):
        tableheader_fontweight = kwargs.get("tableheader_fontweight", "bold")
        tableheader_fontsize = kwargs.get("fontsize", 12)
        nlast = len(ax.get_yticklabels())  # last row is table header
        ax.get_yticklabels()[nlast - 1].set_fontweight(tableheader_fontweight)
        ax.get_yticklabels()[nlast - 1].set_fontsize(tableheader_fontsize)
    return ax


def format_xlabel(xlabel: str, ax: Axes, **kwargs: Any) -> Axes:
    """
    Format the x-axis label.

    Parameters
    ----------
    xlabel (str)
            Title of the left-hand side x-axis.
    ax (Matplotlib Axes)
            Axes to operate on.

    Returns
    -------
            Matplotlib Axes object.
    """
    if xlabel is not None:
        xlabel_size = kwargs.get("xlabel_size", 12)
        xlabel_fontweight = kwargs.get("xlabel_fontweight", "bold")
        ax.set_xlabel(xlabel, size=xlabel_size, fontweight=xlabel_fontweight)
    return ax


def format_xticks(
    dataframe: pd.core.frame.DataFrame,
    estimate: str,
    ll: str,
    hl: str,
    xticks: Optional[Union[list, range]],
    ax: Axes,
    **kwargs: Any
) -> Axes:
    """
    Format the xtick labels.

    This function sets the range of the x-axis using the lowest value and highest values
    in the confidence interval.
    Sets the xticks according to the user-provided 'xticks' or just use 5 tickers.

    Parameters
    ----------
    dataframe (pandas.core.frame.DataFrame)
            Pandas DataFrame where rows are variables. Columns are variable name, estimates,
            margin of error, etc.
    estimate (str)
            Name of column containing the estimates (e.g. pearson correlation coefficient,
            OR, regression estimates, etc.).
    ll (str)
            Name of column containing the lower limit of the confidence intervals.
            Optional
    hl (str)
            Name of column containing the upper limit of the confidence intervals.
            Optional
    xticks (list-like)
            List of xtickers to print on the x-axis.
    ax (Matplotlib Axes)
            Axes to operate on.

    Returns
    -------
            Matplotlib Axes object.
    """
    nticks = kwargs.get("nticks", 5)
    xtick_size = kwargs.get("xtick_size", 10)
    xticklabels = kwargs.get("xticklabels", None)
    if ll is not None:
        xlowerlimit = dataframe[ll].min()
        xupperlimit = dataframe[hl].max()
    else:
        xlowerlimit = 1.1 * dataframe[estimate].min()
        xupperlimit = 1.1 * dataframe[estimate].max()
    ax.set_xlim(xlowerlimit, xupperlimit)
    if xticks is not None:
        ax.set_xticks(xticks)
        ax.xaxis.set_tick_params(labelsize=xtick_size)
    else:
        ax.xaxis.set_major_locator(plt.MaxNLocator(nticks))
    ax.tick_params(axis="x", labelsize=xtick_size)
    if xticklabels:
        ax.set_xticklabels(xticklabels)
    for xticklab in ax.get_xticklabels():
        xticklab.set_fontfamily("sans-serif")
    return ax


def draw_alt_row_colors(
    dataframe: pd.core.frame.DataFrame,
    groupvar: str,
    annoteheaders: Optional[Union[Sequence[str], None]],
    right_annoteheaders: Optional[Union[Sequence[str], None]],
    ax: Axes,
    **kwargs: Any
) -> Axes:
    """
    Color alternating rows in the plot.

    Colors the even-numbered rows gray unless they are rows that indicate groups.
    Breaks from groups will restart with gray.

    Parameters
    ----------
    dataframe (pandas.core.frame.DataFrame)
            Pandas DataFrame where rows are variables. Columns are variable name, estimates,
            margin of error, etc.
    groupvar (str)
            Name of column containing group of variables.
    annoteheaders (list-like)
            List of table headers to use as column headers for the additional annotations
            on the left-hand side of the plot.
    right_annoteheaders (list-like)
            List of table headers to use as column headers for the additional annotations
            on the right-hand side of the plot.
    ax (Matplotlib Axes)
            Axes to operate on.

    Returns
    -------
            Matplotlib Axes object.
    """
    # Retrieve settings
    row_color = kwargs.get("row_color", "0.5")
    if (annoteheaders is not None) or (right_annoteheaders is not None):
        headers_exist = True
    else:
        headers_exist = False
    yticklabels = ax.get_yticklabels()
    counter = 1
    if groupvar is not None:
        groups = [
            grp_str.strip().lower()
            for grp_str in dataframe[groupvar].unique()
            if isinstance(grp_str, str)
        ]
    else:
        groups = []
    for ix, ticklab in enumerate(yticklabels):
        if headers_exist and (ix == len(yticklabels) - 1):
            break
        labtext = ticklab.get_text()
        if labtext.lower().strip() in groups:
            counter = 2  # reset
        else:  # color if even row
            if counter % 2 == 0:
                ax.axhspan(ix - 0.5, ix + 0.5, color=row_color, alpha=0.08, zorder=0)
            counter += 1
    return ax


def draw_tablelines(
    dataframe: pd.core.frame.DataFrame,
    righttext_width: float,
    pval: str,
    right_annoteheaders: Optional[Union[Sequence[str], None]],
    ax: Axes,
    **kwargs: Any
) -> Axes:
    """
    Plot horizontal lines as table lines.

    Cf. draw_ylabel2 for righttext_width.

    Parameters
    ----------
    dataframe (pandas.core.frame.DataFrame)
            Pandas DataFrame where rows are variables. Columns are variable name, estimates,
            margin of error, etc.
    righttext_width (float)
            x-axis coordinate of the rightmost character of the right-side annotations.
    ax: Axes

    Returns
    -------
            Matplotlib Axes object.
    """
    first_yticklab = ax.get_yaxis().majorTicks[-1]
    bbox_disp = first_yticklab.label.get_window_extent()
    (x0, _), (x1, _) = ax.transData.inverted().transform(bbox_disp)
    upper_lw, lower_lw = 2, 1.3
    nrows = len(dataframe)
    plt.plot(
        [x0, x1], [nrows - 0.4, nrows - 0.4], color="0", linewidth=upper_lw, clip_on=False
    )
    plt.plot(
        [x0, x1], [nrows - 1.45, nrows - 1.45], color="0.5", linewidth=lower_lw, clip_on=False
    )
    if (right_annoteheaders is not None) or (pval is not None):
        extrapad = kwargs.get("extrapad", 0.05)
        x0 = ax.get_xlim()[1] * (1 + extrapad)
        plt.plot(
            [x0, righttext_width],
            [nrows - 0.4, nrows - 0.4],
            color="0",
            linewidth=upper_lw,
            clip_on=False,
        )
        plt.plot(
            [x0, righttext_width],
            [nrows - 1.45, nrows - 1.45],
            color=".5",
            linewidth=lower_lw,
            clip_on=False,
        )
    return ax



"""State version and import user-facing functions."""
VERSION = (0, 3, 1)

__version__ = ".".join(map(str, VERSION))

from forestplot.dataframe_utils import load_data
from forestplot.plot import forestplot



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Setup for PyPI."""
from setuptools import setup

with open("README.md", "r") as fh:
    LONG_DESCRIPTION = fh.read()

install_requires = ["pandas", "numpy", "matplotlib", "matplotlib-inline<=0.1.3"]
setup(
    name="forestplot",
    version="0.3.1",
    license="MIT",
    author="Lucas Shen",
    author_email="lucas@lucasshen.com",
    maintainer="Lucas Shen",
    maintainer_email="lucas@lucasshen.com",
    url="https://github.com/lsys/forestplot",
    description="A Python package to make publication-ready but customizable forest plots.",
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    packages=["forestplot"],
    install_requires=install_requires,
    keywords=[
        "visualization",
        "python",
        "data-science",
        "dataviz",
        "pandas",
        "matplotlib",
        "mpl",
        "forestplot",
        "blobbogram",
    ],
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
)
